/**
 * 
 */
package jp.co.yaskawa.rtc.integratedMotion.module;

import java.util.Arrays;

/**
 * @author motohisa
 *
 */
public class RTMath {

	public static final double YE_M_2PI = 2.0*Math.PI;
	public static final double YE_M_PI = Math.PI;
	public static final double YE_M_PI_2 = Math.PI/2.0;
	public static final double EPSILON = 1E-15;

    private static final double[]  seidot = 
    { 0.0,     /* 0 */
       0.1,     /* 1 */
       0.01,    /* 2 */
       0.001,   /* 3 */
       0.0001,  /* 4 */
       0.00001, /* 5 */
       0.000001,        /* 6 */
       0.0000001,       /* 7 */
       0.00000001,      /* 8 */
       0.000000001,     /* 9 */
       0.0000000001,    /* 10 */
       0.00000000001,   /* 11 */
       0.000000000001,  /* 12 */
       0.0000000000001, /* 13 */
       0.00000000000001,        /* 14 */
       0.000000000000001,       /* 15 */
       0.0000000000000001,      /* 16 */
       0.00000000000000001,     /* 17 */
       0.000000000000000001,    /* 18 */
       0.0000000000000000001,   /* 19 */
       0.00000000000000000001,  /* 20 */
       0.000000000000000000001, /* 21 */
       0.0000000000000000000001,        /* 22 */
       0.00000000000000000000001,       /* 23 */
       0.000000000000000000000001,      /* 24 */
       0.0000000000000000000000001,     /* 25 */
       0.00000000000000000000000001,    /* 26 */
       0.000000000000000000000000001,   /* 27 */
       0.0000000000000000000000000001,  /* 28 */
       0.00000000000000000000000000001, /* 29 */
       0.000000000000000000000000000001,        /* 30 */
       0.0000000000000000000000000000001 };     /* 31 */

	public static int fcmp(double src, double dst, int seido) {
		if(Math.abs((src - dst)) <= seidot[seido])
			return 0;

		return -1;
    }
	
	public static boolean nearlyEqualZero(double dval) {
		//return Math.abs(dval) < epsilon_;
		if (dval == -0.0)
			return true;
		
		if (dval == 0.0)
			return true;

		//return Math.abs(dval) <= seidot[10];
		return Math.abs(dval) <= seidot[3];
	}

	/**
	 * 
	 * @param y
	 * @param x
	 * @return
	 */
	public static double atan3(double y, double x) {
/*		double result;
		if (fcmp(x, 0.0, 10) == 0) {
			if (fcmp(y, 0.0, 10) == 0) {
				return 0.0;
			}
			return (y < 0.0) ? -Math.PI / 2 : Math.PI / 2;
		}
		
		result = Math.atan(y / x);
		if ( x >= 0.0) {
			return result;
		}
		return (y < 0.0) ? result - Math.PI : result + Math.PI;*/
		
		return Math.atan2(y, x);
	}
	
	/**
	 * 引数で与えられたもののうち、最大値を返す.
	 * @param ds
	 * @return
	 */
	public static double max(double...ds) {
	    Arrays.sort(ds);
	    return ds[ds.length - 1];
	}
	
	/**
	 * 引数で与えられたもののうち最小値を返す.
	 * @param ds
	 * @return
	 */
	public static double min(double...ds) {
	    Arrays.sort(ds);
	    return ds[0];
	}
	
	/**
	 * 引数で与えられた値と一致する配列内の値のインデックス値を返す.
	 * 配列内に一致するものが複数ある場合、最初のインデックスを返す.
	 * 一致するものが無い場合、-1を返す.
	 * @param d
	 * @param ds
	 * @return
	 */
	public static int getIndex(double d, double...ds) {
	    int i = 0;
	    for (; i < ds.length; i++) {
	        if (d == ds[i]) {
	            return i;
	        }
	    }
	    return -1;
	}
	
	/**
	 * 与えられた値の合計値を返す。
	 * @param ds
	 * @return
	 */
	public static double sum(double...ds) {
		double sum = 0.0;
		for (double d : ds) {
			sum += d;
		}
		return sum;
	}
	
}
