/**

  COPYRIGHT (c) 2010 Yaskawa Electric Corporation.
  
  $Revision: $
  $Author:   $
  $Date:     $
  
  Description:
  制御点を維持したまま冗長軸を絶対位置指令で動作
  
 */
package jp.co.yaskawa.rtc.integratedMotion.module.motion;

import jp.co.yaskawa.rtc.integratedMotion.controller.FrameType;
import jp.co.yaskawa.rtc.integratedMotion.controller.RightLeft;
import jp.co.yaskawa.rtc.integratedMotion.module.IntegratedMotionInterpolator;
import jp.co.yaskawa.rtc.integratedMotion.module.Parameters;
import jp.co.yaskawa.rtc.integratedMotion.module.RTMath;
import jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext;

/**
 * @author motohisa
 *
 */
public class MoveCPHoldAbs<T> extends AbstractMotion<T> {

	IntegratedMotionInterpolator unit_;
	
	double[] currentRedundantPos_;
	double segmentPeriod_;
	RightLeft rightleft_;         //!< 制御対象が右アームか左アームか
	
	double[] currentPos_;         //!< elbow, lumber1-3, vehicle x,y,theta
	double[] targetPos_;         //!< elbow, lumber1-3, vehicle x,y,theta
	double[] demandVelocity_;
	double[] demandAccel_;
	double targetVelocity_ = 0.0;

	double[] deltaPos_;
	double[] absDeltaPos_;
	int maxAxisNum_;
    double[] ratioArray_;

	double elbow_cmd_;
	double lumbarRx_cmd_;
	double lumbarRy_cmd_;
	double lumbarRz_cmd_;
	double vehicleX_cmd_;
	double vehicleY_cmd_;
	double vehicleR_cmd_;

	double elbow_fb_;
	double lumbarRx_fb_;
	double lumbarRy_fb_;
	double lumbarRz_fb_;
	double vehicleX_fb_;
	double vehicleY_fb_;
	double vehicleR_fb_;

	double delta_elbow_;
	double delta_lumbarRx_;
	double delta_lumbarRy_;
	double delta_lumbarRz_;
	double delta_vehicleX_;
	double delta_vehicleY_;
	double delta_vehicleR_;
	
	double velocity_ = 0.0;
	double accel_ = 0.0;
	
	double[] outputPos_;
	double[] currentPostureR_;
	double[] currentPostureL_;
	double[] currentPosture_ = new double[24];
		
	boolean isMoving_ = false;
	
	public MoveCPHoldAbs(IntegratedMotionInterpolator u, T type) {
		COMMAND = type;
		unit_ = u;
	}
	
	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.AbstractMotion#setParameters(jp.co.yaskawa.rtc.integratedMotion.module.Parameters)
	 */
	@Override
	public void setParameters(Parameters l) {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.DoMotion#doSegment(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public void doSegment(RTEmuContext<T> c) throws MotionAlarmException {

		isMoving_ = true;
				
		velocity_ = velocity_ + accel_ * segmentPeriod_;
		velocity_ = velocity_ > targetVelocity_ ? targetVelocity_ : velocity_;

		double remainPos = Math.abs(targetPos_[maxAxisNum_] - currentPos_[maxAxisNum_]);
		double velocityReducePoint = Math.sqrt( 2 * accel_ * remainPos) - accel_ * segmentPeriod_ * 0.5d;
		velocityReducePoint = velocityReducePoint < 0 ? 0 : velocityReducePoint;
		
		velocity_ = velocity_ > velocityReducePoint ? velocityReducePoint : velocity_;
		
		if (velocity_ * segmentPeriod_ == 0) {
			isMoving_ = false;
			currentPos_= targetPos_;
		} else {
			double incrementPos = 0.0d;
            for (int i = 0; i < currentPos_.length; i++) {
    			incrementPos = velocity_ * ratioArray_[i] * segmentPeriod_;
    			currentPos_[i] = currentPos_[i] + incrementPos;            	
            }
		}
		
		outputPos_ = new double[8];
		outputPos_[0] = currentPos_[4];
		outputPos_[1] = currentPos_[5];
		outputPos_[2] = currentPos_[6];
		outputPos_[3] = currentPos_[1];
		outputPos_[4] = currentPos_[2];
		outputPos_[5] = currentPos_[3];
		
		if(rightleft_ == RightLeft.RIGHT) {
			outputPos_[6] = currentPos_[0];
			outputPos_[7] = currentRedundantPos_[1];
		} else {
			outputPos_[6] = currentRedundantPos_[0];
			outputPos_[7] = currentPos_[0];		
		}
		
		unit_.setRedundantPos(outputPos_);
		unit_.setCartesianPos(currentPosture_);
				
	}

	public boolean isMoving() {
		return isMoving_;
	}

	public void setCommand(UnitCommand<T> c) {
		System.out.println("MoveCPHoldAbs: setCommand");

		segmentPeriod_ = unit_.getSegmentPeriod();
		
		//! 現在の冗長軸位置
		currentRedundantPos_ = unit_.getRedundantPos();
		//! 現在の制御点位置
		currentPostureR_ = unit_.getCartesianFeedbackPos(RightLeft.RIGHT, FrameType.ORIGINAL_FRAME);			
		currentPostureL_ = unit_.getCartesianFeedbackPos(RightLeft.LEFT, FrameType.ORIGINAL_FRAME);
		for(int j=0; j<12; j++) {
			currentPosture_[j] = currentPostureR_[j];
			currentPosture_[j+12] = currentPostureL_[j];
		}
		
		double[] params = c.getParams();

		//! 基準アームは左右どちらか
		if (params[0] >= 0) {
			rightleft_ = RightLeft.RIGHT; 
			elbow_fb_ = currentRedundantPos_[6];
		} else {
			rightleft_ = RightLeft.LEFT;
			elbow_fb_ = currentRedundantPos_[7];
		}
		lumbarRx_fb_ = currentRedundantPos_[3];
		lumbarRy_fb_ = currentRedundantPos_[4];
		lumbarRz_fb_ = currentRedundantPos_[5];
		vehicleX_fb_ = currentRedundantPos_[0];
		vehicleY_fb_ = currentRedundantPos_[1];
		vehicleR_fb_ = currentRedundantPos_[2];

		elbow_cmd_ = params[1];
		lumbarRx_cmd_ = params[2];
		lumbarRy_cmd_ = params[3];
		lumbarRz_cmd_ = params[4];
		vehicleX_cmd_ = params[5];
		vehicleY_cmd_ = params[6];
		vehicleR_cmd_ = params[7];

    	currentPos_ = new double[7];    	   
        targetPos_ = new double[7];        
		        
        currentPos_[0] = elbow_fb_; 
		currentPos_[1] = lumbarRx_fb_;
		currentPos_[2] = lumbarRy_fb_;
		currentPos_[3] = lumbarRz_fb_;
		currentPos_[4] = vehicleX_fb_;
		currentPos_[5] = vehicleY_fb_;
		currentPos_[6] = vehicleR_fb_;
		
		targetPos_[0] = elbow_cmd_;
		targetPos_[1] = lumbarRx_cmd_;
		targetPos_[2] = lumbarRy_cmd_;
		targetPos_[3] = lumbarRz_cmd_;
		targetPos_[4] = vehicleX_cmd_;
		targetPos_[5] = vehicleY_cmd_;
		targetPos_[6] = vehicleR_cmd_;		
		
        deltaPos_ = new double[7];
		absDeltaPos_ = new double[7];
		
		//! 冗長軸の差分値
        deltaPos_[0] = elbow_cmd_ - elbow_fb_; 
        deltaPos_[1] = lumbarRx_cmd_ - lumbarRx_fb_;
        deltaPos_[2] = lumbarRy_cmd_ - lumbarRy_fb_;
        deltaPos_[3] = lumbarRz_cmd_ - lumbarRz_fb_;
        deltaPos_[4] = vehicleX_cmd_ - vehicleX_fb_;
        deltaPos_[5] = vehicleY_cmd_ - vehicleY_fb_;
        deltaPos_[6] = vehicleR_cmd_ - vehicleR_fb_;
       
        for (int i = 0; i < 7; i++) {
            absDeltaPos_[i] = Math.abs(deltaPos_[i]);
        }
        
/*       		
		//! 冗長軸の差分値
		delta_elbow_ = elbow_cmd_ - elbow_fb_; 
		delta_lumbarRx_ = lumbarRx_cmd_ - lumbarRx_fb_;
		delta_lumbarRy_ = lumbarRy_cmd_ - lumbarRy_fb_;
		delta_lumbarRz_ = lumbarRz_cmd_ - lumbarRz_fb_;
		delta_vehicleX_ = vehicleX_cmd_ - vehicleX_fb_;
		delta_vehicleY_ = vehicleY_cmd_ - vehicleY_fb_;
		delta_vehicleR_ = vehicleR_cmd_ - vehicleR_fb_;
*/		
		demandVelocity_ = new double[6];
		for (int i=0; i<demandVelocity_.length; i++) {
			demandVelocity_[i] = params[i+8];
		}

		demandAccel_ = new double[6];
		for (int i=0; i<demandAccel_.length; i++) {
			demandAccel_[i] = params[i+14];
		}		

		double totalTime[] = new double[7];
		
		for (int i=0; i<totalTime.length; i++) {
			
			if((i==5) || (i==6)){
				if(demandVelocity_[i-1] != 0)
					totalTime[i] = absDeltaPos_[i]/demandVelocity_[i-1];
				else
					totalTime[i] = 0.0;					
			}
			else{
				if(demandVelocity_[i] != 0)
					totalTime[i] = absDeltaPos_[i]/demandVelocity_[i];
				else
					totalTime[i] = 0.0;									
			}	
		}

		double maxTime = RTMath.max(totalTime.clone());
        maxAxisNum_ = RTMath.getIndex(maxTime, totalTime);
		
		targetVelocity_ = demandVelocity_[maxAxisNum_];
		accel_ = demandAccel_[maxAxisNum_];

        ratioArray_ = new double[7];
        for (int i = 0; i < ratioArray_.length; i++) {
            ratioArray_[i] = totalTime[i] / totalTime[maxAxisNum_];
            if( deltaPos_[i] <0 )
            {
            	ratioArray_[i] = ratioArray_[i]*(-1);
            }
        }
        
		isMoving_ = true;
		
	}

}
