/**
 * 
 */
package jp.co.yaskawa.rtc.integratedMotion.module.motion;

import jp.co.yaskawa.rtc.integratedMotion.controller.FrameType;
import jp.co.yaskawa.rtc.integratedMotion.controller.RightLeft;
import jp.co.yaskawa.rtc.integratedMotion.module.IntegratedMotionInterpolator;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.kinematics.HgTransMatrix;

/**
 * @author motohisa
 *
 */
public class MoveLinearCartesianRel<T> extends MoveLinearCartesianAbs<T> {

	public MoveLinearCartesianRel(IntegratedMotionInterpolator u, T type) {
		super(u, type);
	}
	
	public void setCommand(UnitCommand<T> c) {
		System.out.println("MoveLinearCartesianRel: setCommand");
		
		elapsedTime_ = 0.0;
		time_ = 0.0;
		status_ = Status.normal;
		segmentPeriod_ = unit_.getSegmentPeriod();
		HgTransMatrix tmpHgMat_ = new HgTransMatrix();
		boolean frameIsBase = true;
		
		double[] params = c.getParams();

		if (params[0] >= 0) {
			rightleft_ = RightLeft.RIGHT; 
		} else {
			rightleft_ = RightLeft.LEFT; 			
		}
				
		tmpHgMat_.m00 = params[1];
		tmpHgMat_.m01 = params[2];
		tmpHgMat_.m02 = params[3];
		tmpHgMat_.m03 = params[4];
		tmpHgMat_.m10 = params[5];
		tmpHgMat_.m11 = params[6];
		tmpHgMat_.m12 = params[7];
		tmpHgMat_.m13 = params[8];
		tmpHgMat_.m20 = params[9];
		tmpHgMat_.m21 = params[10];
		tmpHgMat_.m22 = params[11];
		tmpHgMat_.m23 = params[12];
		tmpHgMat_.m30 = 0.0;
		tmpHgMat_.m31 = 0.0;
		tmpHgMat_.m32 = 0.0;
		tmpHgMat_.m33 = 1.0;
			
		demandTransVelocity_ = params[13];
		demandRotateVelocity_ = params[14];
		demandTransAccel_ = params[15];
		demandRotateAccel_ = params[16];

		if (params[17] >= 0) {
			isHoldOtherArm_ = true; 
		} else {
			isHoldOtherArm_ = false; 
		}

		if (params[18] >= 0) {
			frameIsBase = true;   //!< Base Frame
		} else {
			frameIsBase = false;  //!< Tool Frame
		}

		if(rightleft_.equals(RightLeft.RIGHT)) {
			currentPostureMain_ = unit_.getCartesianFeedbackPos(RightLeft.RIGHT, FrameType.ORIGINAL_FRAME);
			currentPostureOther_ = unit_.getCartesianFeedbackPos(RightLeft.LEFT, FrameType.ORIGINAL_FRAME);
		}
		else {
			currentPostureMain_ = unit_.getCartesianFeedbackPos(RightLeft.LEFT, FrameType.ORIGINAL_FRAME);
			currentPostureOther_ = unit_.getCartesianFeedbackPos(RightLeft.RIGHT, FrameType.ORIGINAL_FRAME);			
		}
		
		
//		currentPostureR_ = unit_.getCartesianFeedbackPos(RightLeft.RIGHT, FrameType.ORIGINAL_FRAME);
//		currentPostureL_ = unit_.getCartesianFeedbackPos(RightLeft.LEFT, FrameType.ORIGINAL_FRAME);

		startHgMat_.m03 = currentPostureMain_[0];
		startHgMat_.m13 = currentPostureMain_[1];
		startHgMat_.m23 = currentPostureMain_[2];
		startHgMat_.m00 = currentPostureMain_[3];
		startHgMat_.m01 = currentPostureMain_[4];
		startHgMat_.m02 = currentPostureMain_[5];
		startHgMat_.m10 = currentPostureMain_[6];
		startHgMat_.m11 = currentPostureMain_[7];
		startHgMat_.m12 = currentPostureMain_[8];
		startHgMat_.m20 = currentPostureMain_[9];
		startHgMat_.m21 = currentPostureMain_[10];
		startHgMat_.m22 = currentPostureMain_[11];
		startHgMat_.m30 = 0.0;
		startHgMat_.m31 = 0.0;
		startHgMat_.m32 = 0.0;
		startHgMat_.m33 = 1.0;

		if (frameIsBase) {
			endHgMat_.mul(tmpHgMat_,startHgMat_); 
		} else {
			endHgMat_.mul(startHgMat_,tmpHgMat_); 			
		}
		
		transData_.computeTranslationData(startHgMat_, endHgMat_);

		setProfile();
        
        isMoving_ = true;
        
        {
        	System.out.println("SegmentPeriod: " + segmentPeriod_);
        	System.out.println("CurrnetPosture");
        	System.out.print("\t");
        	for (int i = 0; i < currentPostureMain_.length; i++) {
        		System.out.print(currentPostureMain_[i] + " ");
        	}
        	System.out.println();
        	System.out.println("Target");
        	System.out.print("\t");
        	for (int i = 0; i < 12; i++) {
        		System.out.print(params[i] + " ");
        	}
        	System.out.println();
        	System.out.println("target velocity: " + demandTransVelocity_ + ":" + demandRotateVelocity_);
        	System.out.println("target acceleration: " + demandTransAccel_ + ":" + demandRotateAccel_);
            System.out.println("total time: " + totalTime_ + "translationDistance: " + translationDistance_ + "rotationDistance: " + rotationDistance_);
        }
		
	}
	
}
