/**
 * 
 */
package jp.co.yaskawa.rtc.integratedMotion.module.motion;

import jp.co.yaskawa.rtc.integratedMotion.controller.UnitType;
import jp.co.yaskawa.rtc.integratedMotion.module.IntegratedMotionInterpolator;
import jp.co.yaskawa.rtc.integratedMotion.module.Parameters;
import jp.co.yaskawa.rtc.integratedMotion.module.RTMath;
import jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext;

/**
 * @author motohisa
 *
 */
public class MoveUnitAbs<T> extends AbstractMotion<T> {

	IntegratedMotionInterpolator unit_;
	Parameters parameters_;

	double segmentPeriod_ = 0.0d;
    boolean isMoving_ = false;
	
	double[] currentAngleArray_;
	double[] targetAngleArray_;
	double[] deltaAngleArray_;
	double[] absDeltaAngleArray_;
    double[] ratioArray_;

    double velocity_ = 0.0d;
    double prevVelocity_ = 0.0d;
    double accel_ = 0.0d;
    double targetAngle_ = 0.0d;
    double targetVelocity_ =0.0d;
    int axisNumMaxDistance_;
    
    UnitType unitType_ = UnitType.RIGHT_ARM;
	
	
	public MoveUnitAbs(IntegratedMotionInterpolator u, T type) {
		COMMAND = type;
		unit_ = u;
	}
	
	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.AbstractMotion#setParameters(jp.co.yaskawa.rtc.integratedMotion.module.Parameters)
	 */
	@Override
	public void setParameters(Parameters l) {
		parameters_ = l;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.DoMotion#doSegment(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public void doSegment(RTEmuContext<T> c) throws MotionAlarmException {

        isMoving_ = true;
        
        velocity_ = velocity_ + accel_ * segmentPeriod_;
        velocity_ = velocity_ > targetVelocity_ ? targetVelocity_ : velocity_;
        
        double remainAngle = Math.abs(targetAngleArray_[axisNumMaxDistance_] - currentAngleArray_[axisNumMaxDistance_]);
        double velocityReducePoint = Math.sqrt(2 * accel_ * remainAngle) - accel_ * segmentPeriod_ * 0.5d;
        velocityReducePoint = velocityReducePoint < 0 ? 0 : velocityReducePoint;
        
        velocity_ = velocity_ > velocityReducePoint ? velocityReducePoint : velocity_;
        
        if (velocity_ * segmentPeriod_ == 0) {
            isMoving_ = false;
            currentAngleArray_ = targetAngleArray_;
        } else {
            double incrementAngle = 0.0d;
            for (int i = 0; i < currentAngleArray_.length; i++) {
                incrementAngle = velocity_ * ratioArray_[i] * segmentPeriod_; //incrementAngle = incrementAngle >= deltaAngleArray_[i] ? deltaAngleArray_[i] : incrementAngle;
                currentAngleArray_[i] = currentAngleArray_[i] + incrementAngle;
                deltaAngleArray_[i] = targetAngleArray_[i] - currentAngleArray_[i];
                absDeltaAngleArray_[i] = Math.abs(deltaAngleArray_[i]);
            }
        }
        
//        for(int i=0; i<currentAngleArray_.length; i++){
//        	System.out.println(currentAngleArray_[i]);
//        }
        
        unit_.setJointCommand(unitType_, currentAngleArray_);

	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.DoMotion#isMoving()
	 */
	public boolean isMoving() {
		return isMoving_;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.DoMotion#setCommand(jp.co.yaskawa.rtc.integratedMotion.module.motion.UnitCommand)
	 */
	public void setCommand(UnitCommand<T> c) {
		
    	System.out.println("MoveUnitAbs: setCommand");

        int jointNum = 7;
        segmentPeriod_ = unit_.getSegmentPeriod();

        double[] params = c.getParams();
        if(params[0] == 1.0) {
        	unitType_ = UnitType.RIGHT_ARM;
        	jointNum = 7;
        }
        if(params[0] == 2.0) {
        	unitType_ = UnitType.LEFT_ARM;
        	jointNum = 7;
        }
        if(params[0] == 3.0) {
        	unitType_ = UnitType.LUMBER;
        	jointNum = 3;
        }
        if(params[0] == 4.0) {
        	unitType_ = UnitType.VEHICLE;
        	jointNum = 3;
        }
        
    	currentAngleArray_ = new double[jointNum];
        unit_.getFeedbackPosJoint(unitType_, currentAngleArray_);
        
        targetAngleArray_ = new double[jointNum];        
        deltaAngleArray_ = new double[jointNum];
        absDeltaAngleArray_ = new double[jointNum];
        
        for (int i = 0; i < jointNum; i++) {
        	targetAngleArray_[i] = params[i+1];
            deltaAngleArray_[i] = targetAngleArray_[i] - currentAngleArray_[i];
            absDeltaAngleArray_[i] = Math.abs(deltaAngleArray_[i]);
        }
        
        targetVelocity_ = params[jointNum+1];
        accel_ = params[jointNum+2];

        double maxAbsoluteAngle = RTMath.max(absDeltaAngleArray_.clone());
        axisNumMaxDistance_ = RTMath.getIndex(maxAbsoluteAngle, absDeltaAngleArray_);
        targetAngle_ = targetAngleArray_[axisNumMaxDistance_];
        
        ratioArray_ = new double[jointNum];
        for (int i = 0; i < ratioArray_.length; i++) {
            ratioArray_[i] = deltaAngleArray_[i] / absDeltaAngleArray_[axisNumMaxDistance_];
        }
        
        isMoving_ = true;
        
        if (true) {
        	System.out.println("SegmentPeriod: " + segmentPeriod_);
        	System.out.println("Current");
        	System.out.print("\t");
        	for (int i = 0; i < currentAngleArray_.length; i++) {
        		System.out.print(currentAngleArray_[i] + " ");
        	}
        	System.out.println();
        	System.out.println("Target");
        	System.out.print("\t");
        	for (int i = 0; i < targetAngleArray_.length; i++) {
        		System.out.print(targetAngleArray_[i] + " ");
        	}
        	System.out.println();
        	System.out.println("target velocity: " + targetVelocity_ + " accel: " + accel_ );
        	System.out.println("MaxAngle: " + maxAbsoluteAngle + " axis index: " + axisNumMaxDistance_);
        }
	}

}
