/**

  COPYRIGHT (c) 2010 Yaskawa Electric Corporation.
  
  $Revision: $
  $Author:   $
  $Date:     $
  
  Description:
  モーション実行中状態クラス
  
 */
package jp.co.yaskawa.rtc.integratedMotion.module.state;

import java.util.HashMap;
import java.util.Map;

import jp.co.yaskawa.rtc.integratedMotion.module.Buff;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.AbstractMotion;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MotionAlarmException;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.UnitCommand;

/**
 * @author motohisa
 *
 */
public class Busy<T> implements RTEmuUnitState<T> {

	protected StateObjectHolder<T> holder_;
	protected Buff<UnitCommand<T>> commandQ_;
	protected Map<T,AbstractMotion<T>> motionMap_ = new HashMap<T,AbstractMotion<T>>();
	protected AbstractMotion<T> motion_;

	/*
	 * コマンドバッファのセット
	 * @param q バッファのオブジェクト
	 */
	public void setQ(Buff<UnitCommand<T>> q) {
		commandQ_ = q;
	}
	
	/*
	 * モーション処理オブジェクトを設定する
	 * @param o モーション処理オブジェクト
	 */
	public void putMotion(AbstractMotion<T> o) {
		motionMap_.put(o.COMMAND, o);
	}
	
	
	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#abort(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean abort(RTEmuContext<T> c) {
		c.changeState(holder_.getReadyStateObject());
		commandQ_.clear();
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#causeAlarm(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean causeAlarm(RTEmuContext<T> c) {
		c.changeState(holder_.getAlarmStateObject());
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#clearAlarm(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean clearAlarm(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#doSegment(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public void doSegment(RTEmuContext<T> c) {
		if (motion_ == null) {
			c.changeState(holder_.getReadyStateObject());
			return;
		}
		try {
			motion_.doSegment(c);
			if (!motion_.isMoving()) {
				c.changeState(holder_.getReadyStateObject());
			}
			
		} catch (MotionAlarmException e) {
			Alarm<T> a = holder_.getAlarmStateObject();
			a.set(e);
			c.changeState(a);
		}
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#pause(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean pause(RTEmuContext<T> c) {
		c.changeState(holder_.getPauseStateObject());
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#resume(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean resume(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setAmpPower(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setAmpPower(RTEmuContext<T> c, boolean b) {
		if(!b) {
			commandQ_.clear();
			c.changeState(holder_.getUnPoweredStateObject());
		}		
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setPower(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setPower(RTEmuContext<T> c, boolean b) {
		if(!b) {
			commandQ_.clear();
			c.changeState(holder_.getOffStateObject());
		}		
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setServo(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setServo(RTEmuContext<T> c, boolean b) {
		if(!b) {
			commandQ_.clear();
			c.changeState(holder_.getPoweredStateObject());
		}		
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setStateObjectHolder(jp.co.yaskawa.rtc.integratedMotion.module.state.StateObjectHolder)
	 */
	public void setStateObjectHolder(StateObjectHolder<T> holder) {
		holder_ = holder;

	}
	
	/*
	 * モーション指令オブジェクトをセットする
	 * @param c モーション指令オブジェクト
	 */
	public void setCommand(UnitCommand<T> c) {
		motion_ = getMotion(c.getCommand());
		if (motion_ == null) {
			System.out.println("Busy: motion command is null.");
			return;
		}
		motion_.setCommand(c);
	}

	/*
	 * モーション指令のマップからモーション処理オブジェクトを取得する
	 * @return モーション処理オブジェクト
	 */
	protected AbstractMotion<T> getMotion(T c) {
		AbstractMotion<T> mObj = null;
		mObj = motionMap_.get(c);
		
		return mObj;
	}

	public String getStateString() {
		return "Busy";
	}
	
}
