/**

  COPYRIGHT (c) 2010 Yaskawa Electric Corporation.
  
  $Revision: $
  $Author:   $
  $Date:     $
  
  Description:
  ステートパターン状態インターフェース
  
 */
package jp.co.yaskawa.rtc.integratedMotion.module.state;

/**
 * @author motohisa
 *
 */
public interface RTEmuUnitState<T> {

	/**
	 * 状態オブジェクト保持オブジェクトのセット
	 * @param holder 状態オブジェクト保持オブジェクト
	 */
	void setStateObjectHolder(StateObjectHolder<T> holder);

	/**
	 * 周期実行処理
	 * @param c ステートコンテキスト
	 */
	void doSegment(RTEmuContext<T> c);
	
	
	/**
	 * アンプパワー処理
	 * @param c ステートコンテキスト
	 * @param b アンプパワーON時 true
	 */
	boolean setAmpPower(RTEmuContext<T> c, boolean b);

	/**
	 * サーボON/OFF処理
	 * @param c ステートコンテキスト
	 * @param b サーボON時true
	 */
	boolean setServo(RTEmuContext<T> c, boolean b);

	/**
	 * 主回路電源ON/OFF処理
	 * @param c ステートコンテキスト
	 * @param b 主回路電源ON時true 
	 */
	boolean setPower(RTEmuContext<T> c, boolean b);

	/**
	 * モーション即停止処理
	 * @param c ステートコンテキスト
	 */
	boolean abort(RTEmuContext<T> c);

	/**
	 * モーション一時停止処理
	 * @param c ステートコンテキスト
	 */
	boolean pause(RTEmuContext<T> c);

	/**
	 * モーション再開処理
	 * @param c ステートコンテキスト
	 */
	boolean resume(RTEmuContext<T> c);

	/**
	 * アラームクリア処理
	 * @param c ステートコンテキスト
	 */
	boolean clearAlarm(RTEmuContext<T> c);

	/**
	 * アラーム発生
	 * @param c ステートコンテキスト
	 */
	boolean causeAlarm(RTEmuContext<T> c);
	
	/**
	 * 状態文字列を取得する
	 */
	String getStateString();
	
}
