/**

  COPYRIGHT (c) 2010 Yaskawa Electric Corporation.
  
  $Revision: $
  $Author:   $
  $Date:     $
  
  Description:
  レディ状態クラス
  
 */
package jp.co.yaskawa.rtc.integratedMotion.module.state;

import jp.co.yaskawa.rtc.integratedMotion.module.Buff;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.UnitCommand;

/**
 * @author motohisa
 *
 */
public class Ready<T> implements RTEmuUnitState<T> {

	protected StateObjectHolder<T> holder_;
	protected Buff<UnitCommand<T>> commandQ_;
	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#abort(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean abort(RTEmuContext<T> c) {
		commandQ_.clear();
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#causeAlarm(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean causeAlarm(RTEmuContext<T> c) {
		c.changeState(holder_.getAlarmStateObject());
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#clearAlarm(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean clearAlarm(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#doSegment(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public void doSegment(RTEmuContext<T> c) {
		UnitCommand<T> cmd;
		cmd = commandQ_.get();
		if (cmd != null) {
			Busy<T> busy = holder_.getBusyStateObject();
			
			System.out.println("Ready: setCommand");
			
			busy.setCommand(cmd);
			c.changeState(busy);
			busy.doSegment(c);
		}
		
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#pause(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean pause(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#resume(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean resume(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setAmpPower(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setAmpPower(RTEmuContext<T> c, boolean b) {
		if(!b) {
			commandQ_.clear();
			c.changeState(holder_.getUnPoweredStateObject());
		}		
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setPower(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setPower(RTEmuContext<T> c, boolean b) {
		if(!b) {
			commandQ_.clear();
			c.changeState(holder_.getOffStateObject());
		}		
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setServo(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setServo(RTEmuContext<T> c, boolean b) {
		if(!b) {
			commandQ_.clear();
			c.changeState(holder_.getPoweredStateObject());
		}				
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setStateObjectHolder(jp.co.yaskawa.rtc.integratedMotion.module.state.StateObjectHolder)
	 */
	public void setStateObjectHolder(StateObjectHolder<T> holder) {
		holder_ = holder;

	}

	/*
	 * コマンドバッファのセット
	 * @param q バッファのオブジェクト
	 */
	public void setQ(Buff<UnitCommand<T>> q){
		commandQ_ = q;
	}

	public String getStateString() {
		return "Ready";
	}


}
